@@ -15,7 +15,6 @@
 #include <zlib.h>
 #include <fcntl.h>
 
-
 Bool JoinCTL ();
 void JoinVNC ();
 int CreateMultiSockToRecv (char *, short);
@@ -139,7 +138,6 @@ static struct jpeg_source_mgr jpegSrcManager;
 static JOCTET *jpegBufferPtr;
 static size_t jpegBufferLen;
 
-
 static void
 JpegInitSource(j_decompress_ptr cinfo)
 {
@@ -290,6 +288,8 @@ JpegSetSrcManager(j_decompress_ptr cinfo, CARD8 *compressedData,
 
 #define OPTNUM 2
 
+char interface_name[IF_LENGTH];
+char interface_addr[IF_LENGTH] = "0.0.0.0";
 
 int client_main(int argc, char **argv)
 {   
@@ -297,6 +297,13 @@ int client_main(int argc, char **argv)
   int c, option_index;
   struct option *options;
   
+#if IMAGE_PROC
+  int s_flag = 0;	/* server's client or not */
+
+  /* thread mutex initialize */
+  pthread_mutex_init (&i_mutex, NULL);
+#endif
+
   options = malloc(sizeof(struct option) * OPTNUM);
   options[0].name = "nonotice";
   options[0].has_arg = no_argument;
@@ -309,7 +316,11 @@ int client_main(int argc, char **argv)
 
   /* command line check */
   programName = argv[0];
-  while ((c = getopt_long(argc, argv, "d:s:u:fh", options, &option_index)) != EOF) {
+#if IMAGE_PROC
+  while ((c = getopt_long(argc, argv, "d:s:u:i:fqh", options, &option_index)) != EOF) {
+#else
+  while ((c = getopt_long(argc, argv, "d:s:u:i:fh", options, &option_index)) != EOF) {
+#endif
 //  while ((c = getopt(argc, argv, "d:u:fnh")) != EOF) {
     switch (c) {
       case 1:
@@ -326,10 +337,19 @@ int client_main(int argc, char **argv)
 	memcpy((char *)client_CB.client_ID, optarg, IDLENGTH);
 	auto_flag = 1;
 	break;
+      case 'i':
+	memcpy((char *)interface_name, optarg, IF_LENGTH);
+    get_ifaddr(interface_name, interface_addr);
+	break;
       case 'p':
 	passwd_flag = 1;
 //	memcpy((char *)client_CB.passwd, optarg, IDLENGTH);
 	break;
+#if IMAGE_PROC
+      case 'q':
+	s_flag = 1;
+	break;
+#endif
       case 'h':
       case '?':
       default:
@@ -356,6 +376,13 @@ int client_main(int argc, char **argv)
  				      (GdkInputFunction)GetMultiVNCServerMsg, \
 				      (gpointer)client_CB.mltdsp);
   
+#if IMAGE_PROC
+  if (!s_flag) {
+      /* Create thread */
+      pthread_create (&i_thread, NULL, (void *)image_proc_client, (void *)&client_CB);
+  }
+#endif
+
   while (1) {
     gtk_main_iteration();
     usleep(1);
@@ -397,6 +424,12 @@ JoinCTL ()
   struct sockaddr_in dest;
   ClientAdv clientadv;
 
+#if 0
+  // randam sleep (sleep:0 ~ 1.9s, step:0.05,)
+  sleep(rand() %2);
+  usleep((rand() %20) * 50000);
+#endif
+
   /* Open socket for client control */
   if ( (client_CB.ctldsp = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) {
     fprintf(stderr, "%s : Socket\n", programName);
@@ -425,41 +458,65 @@ JoinCTL ()
   return True;
 }
 
+void
+randam_sleep () {
+  // randam sleep (sleep:0 ~ 5.5s, step:0.5s)
+  sleep(rand() %6);
+  usleep((rand() %2) * 500000);
+}
+
+int
+print_popen(FILE *fp) {
+  char buf[MAXBUF];
+
+  fgets(buf, MAXBUF -1 , fp);
+  fprintf(stderr,"print_popen : %u\n", atoi(buf)); 
+
+  return (atoi(buf) == 0) ? 0 : 1;
+
+}
+
 void 
 JoinVNC () {
-  char commandvnc1[256] = "vncconfig -display :";
-  char commandvnc2[256] = "vncconnect -display :";
-  char commandvnc3[256] = "vnc4config -display :";
+  char commandvnc1[MAXBUF];
+  char commandvnc2[MAXBUF];
+  char commandvnc3[MAXBUF];
   char serverport[10];
   int errno;
+  int count;
+  FILE *fp;
   
-  /* make sh command */
   fprintf(stderr, "vncdisplay=%s\n",vncdisplay);
-  strcat (commandvnc1, vncdisplay);
-  strcat (commandvnc2, vncdisplay);
-  strcat (commandvnc3, vncdisplay);
-  strcat (commandvnc1, " -connect ");
-  strcat (commandvnc2, " ");
-  strcat (commandvnc3, " -connect ");
-  strcat (commandvnc1, (char *)inet_ntoa(client_CB.server_ip));
-  strcat (commandvnc2, (char *)inet_ntoa(client_CB.server_ip));
-  strcat (commandvnc3, (char *)inet_ntoa(client_CB.server_ip));
-  strcat (commandvnc1, ":");
-  strcat (commandvnc2, ":");
-  strcat (commandvnc3, ":");
   sprintf (serverport, "%d\0", ntohs(client_CB.server_port));
-  strcat (commandvnc1, serverport);
-  strcat (commandvnc2, serverport);
-  strcat (commandvnc3, serverport);
+  sprintf (commandvnc1, "vncconfig -display :%s -connect %s:%s > /dev/null 2>&1; echo -n $?", \
+	   vncdisplay, (char *)inet_ntoa(client_CB.server_ip), serverport);
+  sprintf (commandvnc2, "vncconnect -display :%s %s:%s > /dev/null 2>&1; echo -n $?", \
+	   vncdisplay, (char *)inet_ntoa(client_CB.server_ip), serverport);
+  sprintf (commandvnc3, "vnc4config -display :%s -connect %s:%s > /dev/null 2>&1; echo -n $?", \
+	   vncdisplay, (char *)inet_ntoa(client_CB.server_ip), serverport);
+
+
+  /* made sh command */
   fflush (stdout);
   fprintf (stdout, "%s\n", commandvnc1);
   fprintf (stdout, "%s\n", commandvnc2);
   fprintf (stdout, "%s\n", commandvnc3);
 
   /* execute vncconnect */
-  client_CB.vncdsp = (FILE*)system(commandvnc1);
-  client_CB.vncdsp = (FILE*)system(commandvnc2);
-  client_CB.vncdsp = (FILE*)system(commandvnc3);
+  for (count=0; count < TRYCOUNT; count++) {
+      fp = popen(commandvnc1, "r");
+      if (print_popen(fp) == 0) break;
+      fp = popen(commandvnc2, "r");
+      if (print_popen(fp) == 0) break;
+      fp = popen(commandvnc3, "r");
+      if (print_popen(fp) == 0) break;
+
+      randam_sleep ();
+  }
+  fprintf(stderr,"%d\n",count);
+  if (count >= TRYCOUNT) exit (EXIT_FAILURE);
+
+  pclose(fp);
 
   client_CB.connectstate = STAT_VNC;
 }
@@ -487,7 +544,16 @@ CreateMultiSockToRecv (char *addr, short port) {
   adv.sin_addr.s_addr = inet_addr(addr);
 
   stMreq.imr_multiaddr.s_addr = inet_addr(addr);
-  stMreq.imr_interface.s_addr = INADDR_ANY;
+//  stMreq.imr_interface.s_addr = INADDR_ANY;
+  stMreq.imr_interface.s_addr = inet_addr(interface_addr);
+
+//  if (setsockopt(sock, IPPROTO_IP, IP_MULTICAST_IF, (char *) &stMreq.imr_interface.s_addr, sizeof(stMreq.imr_interface.s_addr)) < 0) {
+//    fprintf(stderr, "%s : CreateMultiSockToSend : ip_multicast_if\n", programName);
+//    close(sock);
+//    exit (1);
+//  }
+
+
   setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char*) &stMreq, sizeof(stMreq));
 
   bind(sock, (struct sockaddr *) &adv, sizeof(adv));
@@ -539,8 +605,8 @@ MultiRecieve1 ()
       		    (struct sockaddr *)&recvadd, (socklen_t *)&addr_len);
   strncpy(client_CB.server_ID, serveradv.server_ID, IDLENGTH);
 
-  if (strcmp((char *)client_CB.server_ID, (char *)client_CB.client_ID) == 0) {
-//  if (memcmp((char *)client_CB.server_ID, (char *)client_CB.client_ID, IDLENGTH) == 0) {
+//  if (strcmp((char *)client_CB.server_ID, (char *)client_CB.client_ID) == 0) {
+  if (memcmp((char *)client_CB.server_ID, (char *)client_CB.client_ID, IDLENGTH) == 0) {
     client_CB.server_ip = inet_addr(SERVER_IP);
     client_CB.server_port = htons(SERVER_PORT);
   } else {
@@ -552,7 +618,6 @@ MultiRecieve1 ()
   fprintf(stderr, "type=%d\n",serveradv.type);
   fprintf(stderr, "length=%d\n",serveradv.length);
   fprintf(stderr, "serverID=%s\n",serveradv.server_ID);
-  fprintf(stderr, "serverID=%x\n",serveradv.server_ID);
   fprintf(stderr, "serverip=%s\n",(char *)inet_ntoa(recvadd.sin_addr));
   fprintf(stderr, "serverport=%d\n\n",ntohs(serveradv.server_port));
 
@@ -662,8 +727,6 @@ GetMultiVNCServerMsg (void *g_dsp)
         printf("switch : rfbAllowAdv\n");
 	if ((client_CB.connectstate == STAT_ADV) || (client_CB.connectstate == STAT_CUT)) {
 	  if (buf[1] == 1) {
-	      sleep(rand() %6);
-	      usleep((rand() %2) * 500000);
               JoinVNC ();
 	  }
 	  else if (buf[1] == 0) {
@@ -800,6 +863,12 @@ GetMultiVNCServerMsg (void *g_dsp)
         client_CB.connectstate = buf[1];
 	break;
 
+#if IMAGE_PROC
+      case rfbHistogramRequestAdv:
+	HistogramRequest();
+	break;
+#endif
+
       default:
 	fprintf(stderr, "%s : Undefind Message\n", programName);
     } 
@@ -1196,6 +1265,7 @@ help (char *cmd)
   fprintf(stderr, "  -d [num]\t\tDisplay number of vncserver\n\n");
   fprintf(stderr, "  -s [server_name]\tInput server ID\n\n");
   fprintf(stderr, "  -u [user_name]\tInput client ID and auto connect.\n\n");
+  fprintf(stderr, "  -i [interface]\tInput interface device (eth0).\n\n");
   fprintf(stderr, "  --nonotice\t\tState window is not displayed.\n\n");
 }
 
